/* Written by Joe Freeman 12/90
 *
 * This class uses one external nib file to manage the face display
 * The face panel is loaded with this as the owner and then this class
 * manages the handling of the panel.  I could have created another class
 * ala ConfChooser, and talked to it via messaging but the extra class
 * seems silly.
 *
 * Note: in the ConfAttend.nib the first face cell (not matrix) is connected
 * to a specific message seperate from the matrix.  This is so that I 
 * don't have to set the target of the our own face.  We will redirect the
 * action later.
 *
 */
/* Generated by Interface Builder */

#import "ConfListener.h"
#import "GBListener.h"		/* <jf> */
#import "GraphicView.h"

#import <appkit/Application.h>	/* to pick up perform:with:afterDelay:...*/
				/* also gets priority levels for listeners */
				/* NXApp NXUserName() */
#import <appkit/ButtonCell.h>	
#import <appkit/Matrix.h>
#import <appkit/NXImage.h>
#import <soundkit/Sound.h>
#import <mach.h>		/* vm page size */
#import <sys/param.h>		/* MAXPATHLEN */
#import <sys/types.h>		/* stat call */
#import <sys/stat.h>		/* stat call */

@implementation ConfListener

+ new
{
    return [self notImplemented:@selector(new)];
}

- init
{
    NXZone *zone;
    int i;
    BOOL	foundPort;
	[super init];
	
	/* first set up the face panel (probably should be inside loop) */
	/* zone stuff lifted straight from DrawApp */
	zone = NXCreateChildZone([self zone],vm_page_size, vm_page_size, YES);
	NXNameZone(zone, "ConfAttend");
	attendPanel = [NXApp loadNibSection:"ConfAttend.nib" 
			owner:self 
			withNames:NO 
			fromZone:zone];
	NXMergeZone(zone);
	
	/* now drop us into the conference */
	listenerObject = [[GBListener allocFromZone:[self zone]] init];
	[listenerObject setDelegate:self];
	i = 1;
	foundPort = NO;
	while (!foundPort){
		sprintf(tempPortName,"GBC%d",i++);
		if (![listenerObject checkInAs: tempPortName]){
			foundPort = YES;
			//fprintf(stderr,
			//	"ConfListener.m: using port named %s\n", 
			//	tempPortName);
		}
	}
	/* enable messages even when in mouse down loops */
	[listenerObject setPriority:NX_MODALRESPTHRESHOLD];
	[listenerObject addPort];

	/* should put myself into the conference panel at this point 
	 * I decided to do this when we get our conf ID so we can do tag and
	 * picture all at once 
	 */
	[attendPanel orderFront:self];
	return self;	
}

- free
{
	[listenerObject free];
	[super free];
	return nil;
}


/*============================================================
 * housekeeping
 *
 *============================================================*/

/* this is set before the confernce ID */
- setGraphicView:anObject;
{
	graphicView = anObject;
	return self;
}

/* normally set after the graphicView bacause it comes from remote server */
- setConferenceID:(int)anID
{
	myID = anID;
	[graphicView setConferenceID:anID];
	[self addFaceWithName:(char *)NXUserName() andID:anID];
	return self;
}

- setConferenceSpeaker:anObject
{
	speakerObject = anObject;
	return self;
}

- (char *)portNameUsed
{
	return tempPortName;
}


/*============================================================
 * utillity methods for the attendPanel
 *
 *============================================================*/

/* received whenever anyone joins the conference including ourselves */
- addFaceWithName: (char *)userName andID:(int)userID
{
	int nRow, nCol;
	id faceObject;
	char iconFile[MAXPATHLEN+1];
	struct stat statBuf;
	
	[faceMatrix getNumRows:&nRow numCols:&nCol];
	/* try to create an NXImage with the users face in it */
	sprintf(iconFile,"/LocalLibrary/Images/People/%s.tiff",userName);
	/* first see if there is one with that name allready */
	faceObject  = [NXImage findImageNamed: userName];
	if (!faceObject){
		/* then see if we can get one from images/People */
		faceObject = [[NXImage alloc] init];
		/* I think this is a bug in that it should return NO
		 * if no file exists.  but... it returns YES anyway
		 */
		//if (![faceObject useFromFile:(const char *)iconFile])
		if (stat(iconFile, &statBuf) ==0)
			[faceObject useFromFile:(const char *)iconFile];
		else
			/* last resort, use the outline face */
			[faceObject useFromSection:"genericface"];
		[faceObject setName: (const char *)userName];
	}
	
	if (myID == userID){
		/* we are adding ourselves (first) */
		[[faceMatrix cellAt:0 :0] setTag:userID];
		[[faceMatrix cellAt:0 :0] setTitle:userName];	/*makes copy*/
		if (faceObject)
			[[faceMatrix cellAt:0 :0] setIcon: userName];
		[[ideaMatrix cellAt:0 :0] setTag:userID];
		[[markMatrix cellAt:0 :0] setTag:userID];
	} else {
		/* add some other user at end */
		[faceMatrix addCol];	/* these do not cause a display */
		[ideaMatrix addCol];
		[markMatrix addCol];
		[[faceMatrix cellAt:0 :nCol] setTag:userID];
		[[ideaMatrix cellAt:0 :nCol] setTag:userID];
		[[markMatrix cellAt:0 :nCol] setTag:userID];
		[[faceMatrix cellAt:0 :nCol] setTitle:userName];/*makes copy*/
		if (faceObject)
			[[faceMatrix cellAt:0 :nCol] setIcon: userName];
		[faceMatrix sizeToCells];
		[ideaMatrix sizeToCells];
		[markMatrix sizeToCells];
		[[faceMatrix superview] sizeBy:68.0 :0.0];
		[faceMatrix display];
		[ideaMatrix display];
		[markMatrix display];
	}
	return self;
}

/*============================================================
 * Messages from the attendPanel
 *
 * the if statements in this here to protect us in case the button
 * enables are too permissive
 *============================================================*/

/* shouldn't be able to request the marker if we allready have it */
- requestMarker:sender
{
    if ([markMatrix selectedTag] != myID){
	[speakerObject gbUserRequestMarker:myID];
	[[ideaMatrix findCellWithTag:myID] setState:YES];
	[ideaMatrix display];
	NXPing();
    }
    return self;
}

- dropMarker:sender
{
    /* could turn off marker  but someone else will get it soon */
    if ([markMatrix selectedTag] == myID){
	[speakerObject gbUserGivesUpMarker:myID];
    }
    return self;
}

/* note that the matrix sends its message to here 
 * while our face sends its message somewhere else 
 * after getting the guy inserted in the front, drop the marker
 */
- giveMarkerTo:sender
{
    if ([markMatrix selectedTag] == myID){
	[speakerObject gbUser:myID giveMarkerTo:[sender selectedTag]];
	[self dropMarker:sender];
    }
    return self;
}

/*============================================================
 * Macros
 *
 *============================================================*/

#define NOFORWARD()	[graphicView setConferenceSpeaker: nil];
#define	YESFORWARD()	[graphicView setConferenceSpeaker: speakerObject];

/*============================================================
 * deferred routines (to get around single threaded server) 
 *
 *============================================================*/

/* generates outbound archive */
- delayedSendServerAll:sender
{
    [graphicView sendServerAll];
    return self;
}
/*============================================================
 *
 *
 *============================================================*/

/* received whenever a user joins the conference */
-(int)gbUserAdded : (int) uNum
	userName : (char *) uName
	onHost : (char *) hName
{
    
    //fprintf(stderr,"*");
    //fprintf(stderr,"\nUser:%s added from host %s with id %d\n",
    //				uName,hName,uNum);
    [self addFaceWithName:(char *)uName andID:uNum];
    return 0;
}

/* received whenever a user drops out */
-(int)gbDropMe : (int) uNum
{
    
    //fprintf(stderr,"*");
    //fprintf(stderr,"\nUser:%d dropped\n",uNum);
    [[faceMatrix findCellWithTag:uNum] setEnabled:NO];
    return 0;
}

-(int)gbSendServerAll
{
    
    //fprintf(stderr,"*");
    //[[self class] perform:@selector(delayedSendServerAll:)
    [self  perform:@selector(delayedSendServerAll:)
    		with:self
//		afterDelay:1000		/* millisecs */
		afterDelay:100
		cancelPrevious:YES];
    return 0;
}

/* inbound replacing of glist and slist */
-(int)gbUser : (int) myNum
	allObjectsInArchive : (char *) data
	length : (int) len
{
    
    id newList;
    
    NOFORWARD();
    //fprintf(stderr,"*");
    newList = NXReadObjectFromBuffer (data, len);
    [graphicView replaceGlistWith:newList];
    YESFORWARD();
    return 0;
}

-(int)gbUserRequestMarker : (int) uNum
{
    static id soundObj;
    if (!soundObj)
    	 soundObj = [Sound findSoundFor:"PleasePassMarker"];
    //fprintf(stderr,"*");
    //fprintf(stderr,"User: %d requested marker\n",uNum);
    [[ideaMatrix findCellWithTag:uNum] setState:YES];
    [ideaMatrix display];
    NXPing();
    [soundObj play:self]; 
    return 0;
}

-(int)gbUser : (int) myNum
	giveMarkerTo : (int) uNum
{
    
    //fprintf(stderr,"*");
    //fprintf(stderr,"User %d has been given marker\n",uNum);
    if (uNum == myID){ /* we are getting control of the marker */
    	[requestButton setEnabled:NO];
    	[dropButton setEnabled:YES];
	[graphicView setUserActionEnabled:YES];
	/* make our face drop marker */
	[[faceMatrix cellAt:0 :0] setAction:@selector(dropMarker:)];
    } else {	/* someone else has the marker */
    	[requestButton setEnabled:YES];
    	[dropButton setEnabled:NO];
	[graphicView setUserActionEnabled:NO];
	/* make our face request marker */
	[[faceMatrix cellAt:0 :0] setAction:@selector(requestMarker:)];
    }
    [markMatrix selectCellWithTag:uNum];
    [[ideaMatrix findCellWithTag:uNum] setState:NO];
    [ideaMatrix display];
    NXPing();
    return 0;
}

/* not clear what we should do here since this should immediately be followed
 * by a gbUser:giveMarkerTo: message
 */
-(int)gbUserGivesUpMarker : (int) uNum
{
    fprintf(stderr,"+");
    //fprintf(stderr,"\n,User:%d\n gave up marker\n",uNum);
    return 0;
}

-(int)gbUser : (int) myNum
	addObjectFromArchive : (char *) data
	length : (int) len
{
    
    fprintf(stderr,"+");
    return 0;
}

-(int)gbUser : (int) myNum
	insertObjectFromArchive : (char *) data
	length : (int) len
	at : (int) listNum
{
    id newObj;
    
    NOFORWARD();
    //fprintf(stderr,"*");
    newObj = NXReadObjectFromBuffer (data, len);
    [graphicView insertGraphic:newObj at:listNum];
    YESFORWARD();
    return 0;
}

-(int)gbUser : (int) myNum
	removeObjectAt : (int) listNum
{
    NOFORWARD();
    //fprintf(stderr,"*");
    [graphicView removeGraphicAt: listNum];
    YESFORWARD();
    return 0;
}

-(int)gbUser : (int) myNum
	replaceObjectAt : (int) listNum
	withArchive : (char *) data
	length : (int) len
{
    
    id newObj;
    
    NOFORWARD();
    //fprintf(stderr,"*");
    newObj = NXReadObjectFromBuffer (data, len);
    [graphicView replaceGraphicAt:listNum with:newObj];
    YESFORWARD();
    return 0;
}

-(int)gbUser : (int) myNum
	moveObjectFrom : (int) listNum
	toPosition : (int) destNum
{
    
    fprintf(stderr,"+");
    return 0;
}

-(int)gbUserRecacheSelection : (int) myNum
{
    NOFORWARD();
    //fprintf(stderr,"*");
    [graphicView recacheSelection];
    YESFORWARD();
    return 0;
}

-(int)gbUserGetSelection : (int) myNum
{
    
    NOFORWARD();
    //fprintf(stderr,"*");
    [graphicView getSelection];
    YESFORWARD();
    return 0;
}

-(int)gbUserCacheSelection : (int) myNum
{
    
    NOFORWARD();
    //fprintf(stderr,"*");
    [graphicView cacheSelection];
    YESFORWARD();
    return 0;
}

-(int)gbUserCacheGraphic : (int) myNum
{
    
    fprintf(stderr,"+");
    return 0;
}

-(int)gbUser : (int) myNum
	insertGraphic : (char *) data
	length : (int) len
{
    id newObj;
    
    NOFORWARD();
    //fprintf(stderr,"*");
    newObj = NXReadObjectFromBuffer (data, len);
    [graphicView insertGraphic:newObj];
    YESFORWARD();
    return 0;
}

-(int)gbUser : (int) myNum
	selectObjectAt : (int) listNum
{
    
    NOFORWARD();
    //fprintf(stderr,"*");
    //fprintf(stderr,"received selectObjectAt:%d from %d \n", myNum, listNum);
    [graphicView selectObjectAt:listNum];
    YESFORWARD();
    return 0;
}

-(int)gbUser : (int) myNum
	deselectObjectAt : (int) listNum
{
    
    NOFORWARD();
    //fprintf(stderr,"*");
    [graphicView deselectObjectAt:listNum];
    YESFORWARD();
    return 0;
}

-(int)gbUser : (int) myNum
	changeBounds : (int) listNum
	to : (char *) data
	length : (int) len;
{
    NOFORWARD();
    //fprintf(stderr,"*");
    [graphicView updateBoundsOfGraphic:listNum to:(NXRect *)data];
    YESFORWARD();
    return 0;
}

- gbUserDelete:(int)myNum
{
    
    NOFORWARD();
    //fprintf(stderr,"*");
    [graphicView delete:self];
    YESFORWARD();
    return 0;
}

-(int)gbUserSelectAll : (int) myNum
{
    
    NOFORWARD();
    //fprintf(stderr,"*");
    [graphicView selectAll:self];
    YESFORWARD();
    return 0;
}

-(int)gbUserDeselectAll : (int) myNum
{
    
    NOFORWARD();
    //fprintf(stderr,"*");
    [graphicView deselectAll:self];
    YESFORWARD();
    return 0;
}

-(int)gbUserBringToFront : (int) myNum
{
    
    NOFORWARD();
    //fprintf(stderr,"*");
    [graphicView bringToFront:self];
    YESFORWARD();
    return 0;
}

-(int)gbUserSendToBack : (int) myNum
{
    
    NOFORWARD();
    //fprintf(stderr,"*");
    [graphicView sendToBack:self];
    YESFORWARD();
    return 0;
}

-(int)gbUserGroup : (int) myNum
{
    
    NOFORWARD();
    //fprintf(stderr,"*");
    [graphicView group:self];
    YESFORWARD();
    return 0;
}

-(int)gbUserUnGroup : (int) myNum
{
    
    NOFORWARD();
    //fprintf(stderr,"*");
    [graphicView ungroup:self];
    YESFORWARD();
    return 0;
}

-(int)gbUser : (int) myNum
	graphicsPerform : (int) aSelector
	andDraw : (int) flag
{
    
    NOFORWARD();
    //fprintf(stderr,"*");
    [graphicView graphicsPerform:(SEL)aSelector andDraw:flag];
    YESFORWARD();
    return 0;
}

-(int)gbUser : (int) myNum
	graphicsPerform : (int) aSelector
	with : (char *) argument
	length : (int) len
	andDraw : (int) flag
{
    int	*tempArg;
    NOFORWARD();
    //fprintf(stderr,"*");
    /* the argument for these needs to be an int when passed to perform */
    if (	(SEL)aSelector == @selector(setFill:) ||
    		(SEL)aSelector == @selector(setFramed:) ||
    		(SEL)aSelector == @selector(setLineCap:) ||
    		(SEL)aSelector == @selector(setLineArrow:) ||
    		(SEL)aSelector == @selector(setLineJoin:) ){
	    tempArg = (int *)argument;
	    [graphicView 
			graphicsPerform:(SEL)aSelector 
			with:(void *) *tempArg andDraw:flag];
    }else{
	    [graphicView 
			graphicsPerform:(SEL)aSelector 
			with:argument andDraw:flag];
    }
    if ((SEL)aSelector == @selector(moveBy:)){
    	/* this is just about the only real kludge in this class */
	/* drop this and you won't see moves, unless the user clicks */
	[graphicView graphicsPerform:@selector(activate) andDraw:YES];
    }
    /* this needs to be done for moveBy: and changing linewidths... */
    [[graphicView window] flushWindow];
    YESFORWARD();
    return 0;
}



@end
