/*	Written By Joe Freeman
 *
 * potential trap if marker user drops out 
 * should notify everyone when someone doesn't respond 
 *		(potential deadlock if do though)
 * needs heartbeat on user with mouse
 *
 * this is the server.  It handles 
 * conference attendees adding and dropping.
 * marker arbitration
 * message echoing (so that all users see the same picture
 *
 * There are several methods that are invoked after a delay
 *	delayedUpdateUser
 *	delayedGiveOutMarker
 *	gbSendServerAll
 * These are used so that server can return from the speaker listener message
 * allowing the sender of a message to get back to it's event loop.  
 * Then the server can send a message to anyone it wants, including the
 * originator of message that caused all this.  Otherwise, we have a deadlock
 * where the client program sends a message to the server, hanging untill 
 * return, while the server tries to send a message to the client in reply,
 * waiting for a response.  They will eventually time out but...
 *
 *
 * Log
 * 12/??	started
 * 12/23	added marker arbitration including delayed methods
 * 12/24	added in code to recover from users that vanish
 * 1/2/91	added more flexible marker arbitration 
 * 
 */
/* Generated by Interface Builder */

#import "GBServer.h"
#import "GBListener.h"
#import "GBSpeaker.h"

#import <c.h>
#import <appkit/Application.h>
#import <appkit/Button.h>
#import <objc/HashTable.h>
#import <objc/List.h>
#import <appkit/nextstd.h>		/* NXMalloc */
#import <appkit/Panel.h>
#import <appkit/ScrollView.h>
#import <appkit/Text.h>

@implementation GBServer

/*============================================================
 *
 *
 *============================================================*/

- setEntriesSwitch:anObject
{
    entriesSwitch = anObject;
    loggingEntries = YES;
    [entriesSwitch setState:YES];
    return self;
}

- setLogScroller:anObject
{
    logScroller=anObject;
    logText = [logScroller docView];
    return self;
}

- logString:(char *)theString
{
    [logText setSel: [logText textLength] :[logText textLength]];
    [logText replaceSel: theString];
    [logText scrollSelToVisible];
    return self;
}

- logString:(char *)theString withUser:(int)userNum
{
    char tempString[255];
    sprintf(tempString, theString, userNum);
    [self logString:tempString];
    return self;
}

#define logGraphics(s,n) if (loggingGraphics)	[self logString:s withUser:n];
#define logEntries(s,n)	 if (loggingEntries)	[self logString:s withUser:n];
#define logDraws(s,n)	 if (loggingDraws) 	[self logString:s withUser:n];
#define logAdmit(s,n)	 if (loggingAdmit) 	[self logString:s withUser:n];

/*============================================================
 *
 *
 *============================================================*/

- appDidInit:sender
{
    confListener= [[GBListener  allocFromZone:[self zone]] init];
    [confListener setDelegate: self];
    if ([confListener checkInAs:"GBS"]){
    	NXRunAlertPanel("GBS","Can't check in.\n'","Die",NULL,NULL);
	[NXApp terminate:self];
    }
    [confListener addPort];
    
    speakerList = 	[[List alloc] init];
    deadList = 		[[List alloc] init];
    confNames = 	[HashTable newKeyDesc:"i" valueDesc: "*"];
    confHosts = 	[HashTable newKeyDesc:"i" valueDesc: "*"];
    wantMarkerList = 	[[List alloc] init];

    [self logString:"GreyBoardServer Ports Opened\n"];
    
    return self;
}

- free
{
    [confListener free];
    [speakerList freeObjects];
    [speakerList free];
    [deadList freeObjects];
    [deadList free];
    [confNames free];
    [confHosts free];
    [wantMarkerList free];
    return self;
}
/*============================================================
 *
 *
 *============================================================*/

- logGraphic:sender
{
    loggingGraphics = [sender state];
    return self;
}

- logAdmit:sender
{
    loggingAdmit = [sender state];
    if (loggingAdmit)
    	[self logString:"Selective admissions to conference enabled\n"];
    else
	[self logString:"All who wish to attend are welcome\n"];
    return self;
}

- logEntries:sender
{
    loggingEntries = [sender state];
    return self;
}

- logDraws:sender
{
    loggingDraws = [sender state];
    return self;
}

/*============================================================
 *
 * private methods 
 *
 *============================================================*/

- moveObjectAt: (int)index fromList: sList toList: dList
{
	[dList insertObject: [sList removeObjectAt: index] at:0 ];
	return self;
}

/* this is one of the few sections I am not convinced about.  Alot of users
 * getting killed off could really cause a problem.
 * log that the object is not responding
 * after delay (to let out of thread) tell all users it is gone
 * after delay notify everyone that the user has been dropped 
 */
#define objDoesNotRespond(n)	{\
	logEntries("user %d not responding\n",(int)[speakerList objectAt:n]);\
	if ((int)[speakerList objectAt:n] == markerHolder) 		\
		[self  perform:@selector(gbUserGivesUpMarker:) 		\
				with:(id)markerHolder 			\
				afterDelay:50		/* millisecs */ \
				cancelPrevious:YES]; 			\
	[self  perform:@selector(gbDropMe:) 				\
				with:(id)[speakerList objectAt:n]	\
				afterDelay:50		/* millisecs */ \
				cancelPrevious:NO]; 			\
	[self moveObjectAt: n  fromList: speakerList toList: deadList];}

- tellAllUsers:(SEL)aSelector requestedBy:(int)userNum
{
    int i;
    for ( i = 0 ; i < [speakerList count]; i++){
 	if( (int)[speakerList objectAt: i] != userNum){
	    if ([[speakerList objectAt:i] 
	    		perform:aSelector
	    		with:(id)userNum])
		objDoesNotRespond( i);
	}
    }
    return self;
}
 
/*======================*/

/* this updates a (new) user by telling it all the members in the conf and who
 * has the maker.  Used so that new user can get control of thread back after
 * adding itself to the conference 
 */
- delayedUpdateUser:sender
{
    int i;
    id	tmpID;
    char tmpName[20];
    char tmpHost[20];
    for ( i = 0 ; i < [speakerList count]; i++){
	/* tell new user this guy is in the conference */
	tmpID = [speakerList objectAt:i];
	strcpy(tmpName, (char *)[confNames valueForKey:(const void *)tmpID]);
	strcpy(tmpHost, (char *)[confHosts valueForKey:(const void *)tmpID]);
	//fprintf(stderr,"username = %s\n",
	//	(char *)[confNames valueForKey:(const void *)tmpID]);
	if (lastAdded != tmpID)
	    [lastAdded 	gbUserAdded: (int)tmpID
			userName:  tmpName
			onHost:    tmpHost];
    }
    /* inform the new user who has the marker */
    [lastAdded gbUser: (int)0 giveMarkerTo: markerHolder];
    return self;
}

-(int)gbAddUser : (int *) uNum
	userName : (char *) uName
	onHost : (char *) hName
	portNamed : (char *) pName
{
	port_t	newPort;
	char	*user, *host;
	id	newSpeaker;
	
	int 	i;
	 
	char logString[255];
	if (loggingAdmit) {
		/* never supposed ot use the same var for diff stuff but...*/
		sprintf(logString,
		    "May %s on machine %s join the conference?",uName,hName) ;
		if(NXRunAlertPanel("Server Query",logString,"Yes","No",NULL)!=
				NX_ALERTDEFAULT){
		    sprintf(logString,"User %s from %s denied\n",uName,hName);
		    logAdmit(logString,0);
		    *uNum=0;
		    return 0;
		}
	}
	sprintf(logString,"addUser named:%s onHost:%s withPort:%s ", 
			uName,hName,pName);
	logEntries(logString,0);
	
	newSpeaker = [[GBSpeaker allocFromZone:[self zone]] init];
	if (!(newPort = NXPortFromName(pName, hName)) )
		NXRunAlertPanel("GBServer",
		"Cannot connect to port for new user\n","OK",NULL,NULL);
	[newSpeaker setSendPort: newPort];
	NX_MALLOC(user, char, strlen(uName)+1);
	NX_MALLOC(host, char, strlen(hName)+1);
	strcpy(user,uName);
	strcpy(host,hName);
	[speakerList addObject:newSpeaker];	/* adds at end */
	[confNames insertKey: newSpeaker value: user];
	[confHosts insertKey: newSpeaker value: host];
	
	*uNum = (int)newSpeaker;
	logEntries(" given number %d\n",*uNum);
	
	/* if this is the only one in the conf then give them the marker
	 * else, notify everyone that this person has joined and
	 *	tell new user who is in conf
	 */
	if ([speakerList count] == 1){
	 	/* do this delayed so sender can get message it has marker*/
		[self  perform:@selector(delayedGiveOutMarker:)
				with:self
				afterDelay:50		/* millisecs */
				cancelPrevious:YES];
	}else{
		for ( i = 0 ; i < [speakerList count]; i++){
			if( [speakerList objectAt: i] != newSpeaker){
				/* tell other about this one */
				if ([[speakerList objectAt:i] 
					gbUserAdded : (int)newSpeaker
					userName :  uName
					onHost : hName])
				objDoesNotRespond( i);
			}
		}
		[self  perform:@selector(delayedUpdateUser:)
				with:self
				afterDelay:50		/* millisecs */
				cancelPrevious:YES];
	}

	lastAdded = newSpeaker;
	/* tell someone to update the new guy */
 	/* do this only if others are present */
	if ([speakerList indexOf:newSpeaker] != 0){
		/* this will be deferred by receiver so control returns */
		while ([[speakerList objectAt:0] gbSendServerAll]){
			objDoesNotRespond( 0);
		}
	}
	return 0;
}

/* should NEVER be received by server -- sent in response to addUser:*/
-(int)gbUserAdded : (int) uNum
	userName : (char *) uName
	onHost : (char *) hName;
{
	char logString[255];
	sprintf(logString,"userAdded name:%s onHost:%s<ERROR>\n",uName,hName);
	logEntries(logString,0);
	return 0;
}

-(int)gbDropMe : (int) uNum;
{
	logEntries("dropping user %d\n",uNum);
	[speakerList removeObject:(id)uNum];
	[wantMarkerList removeObject:(id)uNum];
	[self tellAllUsers:@selector(gbDropMe:)  requestedBy:uNum];
	if (uNum == (int)markerHolder)
	    [self  perform:@selector(delayedGiveOutMarker:)
			with:self
			afterDelay:50		/* millisecs */
			cancelPrevious:YES];
	/* kill off the server when the last user drops out */
	if (![speakerList count])   [NXApp terminate:self];	
	return 0;
}

/* this message should NEVER be received 
 * used to tell someone to update server so it can update new user*/
-(int)gbSendServerAll;
{
	logEntries("requesting page send <ERROR>\n",0);
	return 0;
}

/* the enclosed archive is all the objects on the display usually in response
 * to gbSendServerAll for a new user
 */
-(int)gbUser : (int) myNum
	allObjectsInArchive : (char *) data
	length : (int) len;
{
	logEntries("%d sent server whole page, per request\n",myNum);
	[lastAdded gbUser: myNum
		allObjectsInArchive: (char *) data
		length: (int) len];
	return 0;
}
	
/*======================================================*/


-(int)gbUserRequestMarker : (int) uNum;
{
	logEntries("%d requested marker\n",uNum);
	/* this little bit also allows a user to have the server ping
	 * the user with the marker... 
	 */
	/* can't request marker if you allready have it */
	////if (uNum != markerHolder){
		[wantMarkerList addObjectIfAbsent:(id)uNum];
		[self tellAllUsers:@selector(gbUserRequestMarker:)  
				requestedBy:uNum];
	////}
	return 0;
}

/* all marker handoffs go through this point, except for marker drops *
 * this gets called with myNum==0 if user just dropped which caused delayed
 * which generated this request
 *
 * this gets called with myNum!=0 if user does direct passoff in which case
 * 	then we just insert the desired one at the head of the queue
 * then the user with marker will still have to drop the marker
 */
-(int)gbUser : (int) myNum
	giveMarkerTo : (int) uNum;
{
    int i;
    logEntries("giving marker to %d\n", uNum);
    
    /* if this is from external then insert the guy at front
     * check markerHolder because otherwise first user will get messed  
     */
    if (markerHolder && myNum == markerHolder){
    	[wantMarkerList removeObject:(id)uNum];
	[wantMarkerList insertObject:(id)uNum at:0];
//	[self  perform:@selector(delayedGiveOutMarker:)
//		with:self
//		afterDelay:50		/* millisecs */
//		cancelPrevious:YES];
    }else if (myNum) {
    	/* you can only give the marker away if you own it */
    }else{
  	  if (!uNum) return 0;	/* probably no one in list */
  	  markerHolder = uNum;
  	  /* remove this guy from the requesting list */
   	 [wantMarkerList removeObject:(id)uNum];
    
    	for ( i = 0 ; i < [speakerList count]; i++){
 		if( (int)[speakerList objectAt: i] != myNum){
		    if ([[speakerList objectAt:i] 
				gbUser : myNum
				giveMarkerTo : uNum])
			objDoesNotRespond( i);
		}
   	 }
    }
    return 0;
}
/*
 * This is generally called by a delayed message.  Usually, we receive a 
 * message that implies marker control will change.  But, we need to return
 * to the sender so it will unlock (quit waiting on speaker/listener).  Then
 * sometime later the server receives this message which causes themarker to
 * be given to someone
 *
 * used when the first person comes into the conference 
 * used when marker holder leaves conference gracefully
 * used when some one gives up the marker (explicitly)
 *
 * could also be used in watchdog routine
 *
 * figure out who gets the marker and send it on 
 */
- delayedGiveOutMarker:sender
{
	if ([wantMarkerList count] == 0)
		/* potential problem if first guy in list has dropped out */
		[self gbUser:0 
			giveMarkerTo: (int)[speakerList objectAt:0]];
	else 
		[self gbUser:0 
			giveMarkerTo: (int)[wantMarkerList objectAt:0]];
	return self;
}

-(int)gbUserGivesUpMarker : (int) uNum
{
    int i;
    /* only the one with marker can give up */
    if (uNum != markerHolder) return 0;	
    logEntries("%d giving up marker\n",uNum);
    markerHolder = 0;		/* update our values */
    for ( i = 0 ; i < [speakerList count]; i++){
	if( (int)[speakerList objectAt: i] != uNum){
		if ([[speakerList objectAt:i] 
					gbUserGivesUpMarker : uNum])
			objDoesNotRespond( i);
	}
    }
    /* don't want to block the server so return then hand out marker */				
    [self  perform:@selector(delayedGiveOutMarker:)
		with:self
		afterDelay:50		/* millisecs */
		cancelPrevious:YES];

    return 0;
}
	
/*======================================================*/

-(int)gbUser : (int) myNum
	addObjectFromArchive : (char *) data
	length : (int) len;
{
    int i;
    logGraphics("%d adding object from archive\n",myNum);
    for ( i = 0 ; i < [speakerList count]; i++){
 	if( (int)[speakerList objectAt: i] != myNum){
	    if ([[speakerList objectAt:i] 
			gbUser : myNum
			addObjectFromArchive : data
			length : len])
		objDoesNotRespond( i);
	}
    }
    return 0;
}

-(int)gbUser : (int) myNum
	insertObjectFromArchive : (char *) data
	length : (int) len
	at : (int) listNum;
{
    int i;
    logGraphics("%d inserting object from archive\n",myNum);
    for ( i = 0 ; i < [speakerList count]; i++){
 	if( (int)[speakerList objectAt: i] != myNum){
	    if ([[speakerList objectAt:i] 
			gbUser : (int) myNum
			insertObjectFromArchive : data
			length : len
			at : listNum])
		objDoesNotRespond( i);
	}
    }
	return 0;
}

-(int)gbUser : (int) myNum
	removeObjectAt : (int) listNum;
{
    int i;
    logGraphics("%d removing object from archive\n",myNum);
    for ( i = 0 ; i < [speakerList count]; i++){
 	if( (int)[speakerList objectAt: i] != myNum){
	    if ([[speakerList objectAt:i] 
			gbUser : myNum
			removeObjectAt : listNum])
		objDoesNotRespond( i);
	}
    }
	return 0;
}

-(int)gbUser : (int) myNum
	replaceObjectAt : (int) listNum
	withArchive : (char *) data
	length : (int) len;
{
    int i;
    logGraphics("%d replacing object from archive\n",myNum);
    for ( i = 0 ; i < [speakerList count]; i++){
 	if( (int)[speakerList objectAt: i] != myNum){
	    if ([[speakerList objectAt:i] 
			gbUser : myNum
			replaceObjectAt : listNum
			withArchive : data
			length : len])
		objDoesNotRespond( i);
	}
    }
	return 0;
}

-(int)gbUser : (int) myNum
	moveObjectFrom : (int) listNum
	toPosition : (int) destNum;
{
    int i;
    logGraphics("%d moving object from archive\n",myNum);
    for ( i = 0 ; i < [speakerList count]; i++){
 	if( (int)[speakerList objectAt: i] != myNum){
	    if ([[speakerList objectAt:i] 
			gbUser : myNum
		moveObjectFrom : listNum
		toPosition : destNum])
		objDoesNotRespond( i);
	}
    }
	return 0;
}

-(int)gbUserRecacheSelection : (int) myNum;
{
	logGraphics("%d recaching selection\n",myNum);
	[self tellAllUsers:@selector(gbUserRecacheSelection:)  requestedBy:myNum];
	return 0;
}

-(int)gbUserGetSelection : (int) myNum;
{
	logGraphics("%d user get selection\n",myNum);
	[self tellAllUsers:@selector(gbUserGetSelection:)  requestedBy:myNum];
	return 0;
}

-(int)gbUserCacheSelection : (int) myNum;
{
	logGraphics("%d caching selection\n",myNum);
	[self tellAllUsers:@selector(gbUserCacheSelection:)  requestedBy:myNum];
	return 0;
}

-(int)gbUserCacheGraphic : (int) myNum;
{
	logGraphics("%d cacheing graphic\n",myNum);
	[self tellAllUsers:@selector(gbUserCacheGraphic:)  requestedBy:myNum];
	return 0;
}

/* conver for insertGraphic:g */
-(int)gbUser : (int) myNum
	insertGraphic : (char *) data
	length : (int) len;
{
    int i;
    logGraphics("%d inserting graphic\n",myNum);
    for ( i = 0 ; i < [speakerList count]; i++){
 	if( (int)[speakerList objectAt: i] != myNum){
	    if ([[speakerList objectAt:i] 
			gbUser : myNum
			insertGraphic : data
			length : len])
		objDoesNotRespond( i);
	}
    }
	return 0;
}

-(int)gbUser : (int) myNum
	selectObjectAt : (int) listNum;
{
    int i;
    logGraphics("%d selecting object at\n",myNum);
    for ( i = 0 ; i < [speakerList count]; i++){
 	if( (int)[speakerList objectAt: i] != myNum){
	    if ([[speakerList objectAt:i] 
			gbUser : (int) myNum
			selectObjectAt : (int) listNum])
		objDoesNotRespond( i);
	}
    }
    return 0;
}

-(int)gbUser : (int) myNum
	deselectObjectAt : (int) listNum;
{
    int i;
    logGraphics("%d deselecting object at\n",myNum);
    for ( i = 0 ; i < [speakerList count]; i++){
 	if( (int)[speakerList objectAt: i] != myNum){
	    if ([[speakerList objectAt:i] 
			gbUser : (int) myNum
			deselectObjectAt : (int) listNum])
		objDoesNotRespond( i);
	}
    }
	return 0;
}

-(int)gbUser : (int) myNum
	changeBounds : (int) listNum
	to : (char *) data
	length : (int) len
{
    int i;
    logGraphics("%d changeBounds\n",myNum);
    for ( i = 0 ; i < [speakerList count]; i++){
 	if( (int)[speakerList objectAt: i] != myNum){
	    if ([[speakerList objectAt:i] 
			gbUser : (int) myNum
			changeBounds : (int) listNum
			to : (char *) data
			length : (int) len])
		objDoesNotRespond( i);
	}
    }
	return 0;
}

- gbUserDelete:(int)myNum;
{
	logGraphics("%d deleting\n",myNum);
	[self tellAllUsers:@selector(gbUserDelete:)  requestedBy:myNum];
	return 0;
}

-(int)gbUserSelectAll : (int) myNum;
{
	logGraphics("%d selecting all\n",myNum);
	[self tellAllUsers:@selector(gbUserSelectAll:)  requestedBy:myNum];
	return 0;
}

-(int)gbUserDeselectAll : (int) myNum;
{
	logGraphics("%d deselecting all\n",myNum);
	[self tellAllUsers:@selector(gbUserDeselectAll:)  requestedBy:myNum];
	return 0;
}

-(int)gbUserBringToFront : (int) myNum;
{
	logGraphics("%d bring to front\n",myNum);
	[self tellAllUsers:@selector(gbUserBringToFront:)  requestedBy:myNum];
	return 0;
}

-(int)gbUserSendToBack : (int) myNum;
{
	logGraphics("%d send to back\n",myNum);
	[self tellAllUsers:@selector(gbUserSendToBack:)  requestedBy:myNum];
	return 0;
}

-(int)gbUserGroup : (int) myNum;
{
	logGraphics("%d group",myNum);
	[self tellAllUsers:@selector(gbUserGroup:)  requestedBy:myNum];
	return 0;
}

-(int)gbUserUnGroup : (int) myNum;
{
	logGraphics("%d ungroup\n",myNum);
	[self tellAllUsers:@selector(gbUserUnGroup:)  requestedBy:myNum];
	return 0;
}

-(int)gbUser : (int) myNum
	graphicsPerform : (int) aSelector
	andDraw : (int) flag;
{
    int i;
    logGraphics("%d graphicsPerform\n",myNum);
    for ( i = 0 ; i < [speakerList count]; i++){
 	if( (int)[speakerList objectAt: i] != myNum){
	    if ([[speakerList objectAt:i] 
			gbUser :  myNum
			graphicsPerform :  aSelector
			andDraw :  flag])
		objDoesNotRespond( i);
	}
    }
	return 0;
}

-(int)gbUser : (int) myNum
	graphicsPerform : (int) aSelector
	with : (char *) argument
	length : (int) len
	andDraw : (int) flag;
{
    int i;
    logGraphics("%d graphicsPerform with\n",myNum);
    for ( i = 0 ; i < [speakerList count]; i++){
 	if( (int)[speakerList objectAt: i] != myNum){
	    if ([[speakerList objectAt:i] 
			gbUser : (int) myNum
			graphicsPerform : (int) aSelector
			with : (char *) argument
			length : (int) len
			andDraw : (int) flag])
		objDoesNotRespond( i);
	}
    }
	return 0;
}



@end
